"""
Class functions.

Usage:
 ./utils/tools/_class_functions.py

Author:
 Peter Rigali - 2022-03-30
"""
from typing import Union


# Model_data
def _names(n: str, n_lst: Union[list, tuple]) -> bool:
    name_dic = {name: True for name in n_lst}
    if n not in name_dic:
        return True
    else:
        raise AttributeError("{} already included in names list".format(n))


def _len(l1: int, l2: int) -> bool:
    if l1 == l2:
        return True
    else:
        raise AttributeError("(len1: {} ,len2: {} ) Lengths are not the same.".format(l1, l2))



"""
Cleaning functions.

Usage:
 ./utils/tools/clean.py

Author:
 Peter Rigali - 2022-03-30
"""
from typing import Union
from collections.abc import KeysView, ValuesView
from pandas import Series, Index
import numpy as np


# Repeat functions
def _mean_(d: Union[list, tuple]) -> float:
    """
    Find the mean value of a list.

    :param d: Input data.
    :type d: list or tuple.
    :return: Mean value.
    :rtype: float.
    :note: *None*
    """
    return sum(d) / d.__len__()


def _variance_(d: Union[list, tuple], ddof: int = 1) -> float:
    """
    Find the variance value of a list.

    :param d: Input data.
    :type d: list or tuple.
    :param ddof: Desired Degrees of Freedom.
    :type ddof: int
    :return: Variance value.
    :rtype: float.
    :note: *None*
    """
    mu = _mean_(d=d)
    return sum(((x - mu) ** 2 for x in d)) / (d.__len__() - ddof)


def _std_(d: list, ddof: int = 1) -> float:
    """
    Find the Standard Deviation value of a list.

    :param d: Input data.
    :type d: list or tuple.
    :param ddof: Desired Degrees of Freedom.
    :type ddof: int
    :return: Standard Deviation value.
    :rtype: float.
    :note: *None*
    """
    return _variance_(d=d, ddof=ddof) ** .5


def _percentile_(d: Union[list, tuple], q: float) -> float:
    """
    Find the percentile value of a list.

    :param d: Input data.
    :type d: list or tuple.
    :param q: Percentile percent.
    :type q: float.
    :return: Percentile value.
    :note: *None*
    """
    d = _round(d=[item * 1000.0 for item in d], v=1)
    ind = _round(d=d.__len__() * q, v=1)
    d.sort()
    for item in d:
        if item >= d[ind]:
            return item / 1000.0


# Misc
def _round(d: Union[list, Series, np.ndarray, float, int], v: float,
           r: bool = False) -> Union[list, float]:
    """
    Rounds a value or list.

    :param d: Value or list.
    :param v: Place to round to.
    :param r: Whether to use remainder. If using floats, this should be true.
    # :param val_type: Desired value type.
    # :type val_type: str.
    :return: Returns a value or list of values.
    :note: *None*
    """
    if isinstance(d, (float, int)):
        if r is True:
            return round(_type(v=d, dtype='float') * v) / v
        else:
            return round(_type(v=d, dtype='float') / v) * v
    elif isinstance(d, (list, Series, np.ndarray)):
        d = (_type(v=i, dtype='float') for i in _mtype(d=d))
        if r is True:
            return [round(item * v) / v for item in d]
        else:
            return [round(item / v) * v for item in d]
    else:
        raise AttributeError('Value not one of the specified types.')


# Cleaning Functions.
def _empty(d) -> bool:
    """Checks if data is empty"""
    if d.__len__() == 0:
        return True
    else:
        return False


def _mtype(d, dtype: str = 'list') -> Union[list, tuple]:
    """Converts list-adjacent objects to a list or tuple"""
    if dtype not in {'list': True, 'tuple': True}:
        raise AttributeError("dtype input must be either {list, tuple}.")
    if isinstance(d, {'list': list, 'tuple': tuple}[dtype]):
        return d
    elif isinstance(d, Series):
        if dtype == 'list':
            return d.to_list()
        else:
            return tuple(d.to_list())
    elif isinstance(d, np.ndarray):
        if dtype == 'list':
            return d.tolist()
        else:
            return tuple(d.tolist())
    elif isinstance(d, (set, KeysView, ValuesView, Index)):
        return {'list': list, 'tuple': tuple}[dtype](d)
    elif isinstance(d, (int, float, str, object, np.int_, np.float_, np.str, np.object)):
        if dtype == 'list':
            return [d]
        else:
            return tuple(d)
    else:
        raise AttributeError('Input data needs to have a type of {np.ndarray, pd.Series, list, set, int, float, str, object}')


def _type(v: Union[float, int, str, object], dtype: str = 'float') -> Union[float, int, str, object]:
    """Converts value to a set type"""
    if isinstance(v, {'float': float, 'int': int, 'str': str, 'object': object}[dtype]):
        return v
    else:
        return {'float': float, 'int': int, 'str': str, 'object': object}[dtype](v)


def _check_type(d: Union[list, tuple], dtype: str = 'float') -> tuple:
    """Checks type of values in a list"""
    return tuple([_type(v=val, dtype=dtype) for val in d])


def _nan(v) -> bool:
    """
    Checks a value to see if Nan.

    :param v: Input value.
    :return: Returns True or False if the value is Nan.
    :rtype: bool.
    :note: *None*
    """
    if v == v and v is not None and v != np.inf and v != -np.inf:
        return False
    else:
        return True


def _rnan(d: Union[list, tuple]) -> list:
    """Remove Nan values from a list"""
    return [val for val in d if _nan(val) is False]


def _rval(d: list, na: str = 'median', std_value: int = 3, cap_zero: bool = True,
          median_value: float = 0.023, ddof: int = 1) -> Union[float, None]:
    """
    Calculate desired replacement for Nan values.

    :param d: Input data.
    :type d: list.
    :param na: Desired Nan value handling method. {zero, mu, std, median}
    :type na: str.
    :param std_value: Desired Standard Deviation to use.
    :type std_value: int.
    :param cap_zero: Whether to cap the value at zero.
    :type cap_zero: bool.
    :param median_value: Desired percentile to use.
    :type median_value: float.
    :return: Replacement value.
    :note: If mean - 3 * std is less than 0, may confuse results.
    """
    if na == 'zero':
        return 0.0
    elif na == 'mu':
        return _mean_(d=_rnan(d=d))
    elif na == 'std':
        d = _rnan(d=d)
        val = _mean_(d=d) - (_std_(d=d, ddof=ddof) * std_value)
        if cap_zero:
            if val > 0:
                return val
            else:
                return 0.0
        else:
            return val
    elif na == 'median':
        return _percentile_(d=_rnan(d=d), q=median_value)
    elif na == 'none':
        return None


def _replace_na(d: Union[list, tuple], rval: float = None) -> Union[list, tuple]:
    """Replace Nan values with replacement value"""
    if rval is None:
        return _rnan(d=d)
    else:
        return tuple([val if _nan(v=val) is False else rval for val in d])


def _prep(d, mtype: str = 'tuple', dtype: str = 'float', na: str = 'zero', std_value: int = 3,
          median_value: float = 0.023, cap_zero: bool = True, ddof: int = 1):
    """Clean data"""
    # Check Empty
    if _empty(d=d):
        raise AttributeError("Data entered is empty.")
    # Convert to list
    d = _mtype(d=d, dtype='list')
    # Check type, check na, replace na
    _na = None
    for ind, val in enumerate(d):
        if _nan(v=val):
            if _na is None:
                _na = _rval(d=d, na=na)
            val = _na
        d[ind] = _type(v=val, dtype=dtype)
    # Convert to metadata
    d = _mtype(d=d, dtype=mtype)
    return d



"""
DataFrame functions.

Usage:
 ./utils/tools/dataframe.py

Author:
 Peter Rigali - 2022-03-30
"""
from typing import Union
from pandas import DataFrame
from pyjr.utils.tools.clean import _mtype


def slc(d: DataFrame, c: str, v: Union[float, int, str, object]) -> DataFrame:
    """
    Select function filters a dataframe using col and a value.

    :param d: Selected DataFrame
    :type d: pd.DataFrame
    :param c: Column
    :type c: str
    :param v: Value to look for.
    :type v: float, int, str, or object
    :return: Returns part of DataFrame matching value within desired column.
    :rtype: pd.DataFrame


    """
    return d[d[c] == v]


def ilc(d: DataFrame, i) -> DataFrame:
    """
    iloc function filters a dataframe using a list of values.

    :param d: Selected DataFrame
    :type d: pd.DataFrame
    :param i: list of indexes to filter rows by.
    :type i: List like objects
    :return: Returns part of DataFrame matching the input index list.
    :rtype: pd.DataFrame


    """
    return d.iloc[_mtype(d=i, dtype='list')]



"""
Cluster class.

Usage:
 ./models/cluster.py

Author:
 Peter Rigali - 2022-03-19
"""
from dataclasses import dataclass
from pyjr.classes.model_data import ModelingData
from sklearn.cluster import KMeans, AffinityPropagation, MeanShift, AgglomerativeClustering, Birch
from sklearn.metrics import rand_score, adjusted_rand_score, mutual_info_score, v_measure_score, silhouette_score
from sklearn.metrics.cluster import contingency_matrix, homogeneity_score, completeness_score


@dataclass
class Cluster:

    __slots__ = ["data", "labels", "pred", "centers", "scores", "name"]

    def __init__(self, data: ModelingData):
        self.data = data
        self.labels = None
        self.pred = None
        self.centers = None
        self.scores = None
        self.name = None

    def __repr__(self):
        return "Cluster"

    def add_kmeans(self, num: int = 8, random_state: int = 0):
        kmeans = KMeans(n_clusters=num, random_state=random_state).fit(self.data.x_train)
        self.labels = kmeans.labels_
        self.pred = kmeans.predict(self.data.x_test)
        self.centers = kmeans.cluster_centers_
        self.name = "CLU_KMeans"
        return self

    def add_affinity(self, random_state: int = 0):
        cluster = AffinityPropagation(random_state=random_state).fit(self.data.x_train)
        self.labels = cluster.labels_
        self.pred = cluster.predict(self.data.x_test)
        self.centers = cluster.cluster_centers_
        self.name = "CLU_AffinityPropagation"
        return self

    def add_mean_shift(self, bandwidth: int = None):
        cluster = MeanShift(bandwidth=bandwidth).fit(self.data.x_train)
        self.labels = cluster.labels_
        self.pred = cluster.predict(self.data.x_test)
        self.centers = cluster.cluster_centers_
        self.name = "CLU_MeanShift"
        return self

    def add_agglomerative(self, num: int = 2):
        cluster = AgglomerativeClustering(n_clusters=num).fit(self.data.x_train)
        self.labels = cluster.labels_
        self.pred = cluster.predict(self.data.x_test)
        self.centers = cluster.cluster_centers_
        self.name = "CLU_AgglomerativeClustering"
        return self

    def add_birch(self, num: int = 2):
        cluster = Birch(n_clusters=num).fit(self.data.x_train)
        self.labels = cluster.labels_
        self.pred = cluster.predict(self.data.x_test)
        self.centers = cluster.subcluster_centers_
        self.name = "CLU_Birch"
        return self

    def add_scores(self):
        self.scores = {"rand": rand_score(labels_true=self.data.y_test, labels_pred=self.pred),
                       "adj rand": adjusted_rand_score(labels_true=self.data.y_test, labels_pred=self.pred),
                       "mutual info": mutual_info_score(labels_true=self.data.y_test, labels_pred=self.pred),
                       "v measure": v_measure_score(labels_true=self.data.y_test, labels_pred=self.pred),
                       "silhouette": silhouette_score(X=self.data.x_test, labels=self.labels),
                       "cont matrix": contingency_matrix(labels_true=self.data.y_test, labels_pred=self.pred),
                       "completeness": completeness_score(labels_true=self.data.y_test, labels_pred=self.pred),
                       "homogenity": homogeneity_score(labels_true=self.data.y_test, labels_pred=self.pred)}
        return self



"""
Stat class.

Usage:
 ./classes/stat.py

Author:
 Peter Rigali - 2022-03-30
"""
from dataclasses import dataclass
from pyjr.utils.tools.math import _min, _max, _mean, _var, _std, _sum, _med, _mod, _skew, _kurt, _perc, _range
from pyjr.utils.tools.clean import _nan, _mtype, _rval, _type, _empty


@dataclass
class Stat:
    """
    The Stat Class allows you to have preset data cleaning functions that are then applied to changing input data.

    :param stat: Desired stat to calculate.
    :type stat: str
    :param na: Desired handing of nan values.
    :type na: str
    :param dtype: Desired output dtype.
    :type dtype: str
    :param empty: If True and the dat is empty, will return 0.
    :type empty: bool
    """
    __slots__ = ('stat', 'na', 'dtype', 'empty')

    def __init__(self, stat: str, na: str = 'zero', dtype: str = 'float', empty: bool = False):
        self.stat = stat
        self.na = na
        self.dtype = dtype
        self.empty = empty

    def get(self, data, q: float = None):
        dic = {'mean': _mean, 'min': _min, 'max': _max, 'var': _var, 'std': _std, 'sum': _sum, 'median': _med,
               'mode': _mod, 'skew': _skew, 'kurt': _kurt, 'percentile': _perc, 'range': _range}
        data = _mtype(d=data, dtype='list')
        if self.empty:
            if _empty(d=data):
                return 0.0
        na = None
        for ind, val in enumerate(data):
            if _nan(v=val):
                if na is None:
                    na = _rval(d=data, na=self.na)
                val = na
            data[ind] = val
        if q is not None and self.stat == 'percentile':
            return _perc(d=data, q=q)
        else:
            return _type(v=dic[self.stat](data=data), dtype=self.dtype)

    def __repr__(self):
        return 'Stat'



"""
Dispatch for tracking function time, memory, and cpu use.

Usage:
 ./utils/evaluate.py

Author:
 Peter Rigali - 2022-03-19
"""
import time
import os
import psutil


def track(func):
    def wrapper(*args, **kwargs):
        cpu_before = psutil.cpu_percent(1)# / os.cpu_count()
        mem_before = psutil.Process(os.getpid()).memory_info().rss

        start = time.time()
        result = func(*args, **kwargs)
        elapsed_time = round(time.time() - start, 2)

        cpu_during = psutil.cpu_percent()# / os.cpu_count()
        mem_after = psutil.Process(os.getpid()).memory_info().rss
        print("{}: mem_consumed: {} runtime: {}, cpu_change: {}".format(func.__name__, mem_after - mem_before,
                                                                        elapsed_time, cpu_during - cpu_before))
        return {"result": result, "memory_before": mem_before, "memory_after": mem_after,
                "execution_time": elapsed_time, "cpu_before": cpu_before, "cpu_during": cpu_during}
    return wrapper

# @track
# def test():
#     for i in range(100):
#         sum([random.randrange(1, 100, 1) for i in range(1000)])
#     return


"""
PreProcess class.

Usage:
 ./utils/preprocess_data.py

Author:
 Peter Rigali - 2022-03-23
"""
from dataclasses import dataclass
from typing import Optional
import math
import numpy as np
from pandas import DataFrame
from pyjr.classes.data import Data
from pyjr.utils.tools.math import _min, _max, _mean, _var, _std, _sum, _med, _mod, _skew, _kurt, _perc
from pyjr.utils.tools.clean import _mtype, _check_type
from sklearn.preprocessing import power_transform, quantile_transform, robust_scale


@dataclass
class PreProcess:

    __slots__ = ("cleanData", "data", "len", "name")

    def __init__(self, data: Optional[Data] = None):
        self.cleanData = data
        self.data = None
        self.len = None
        self.name = None

    def add_normalize(self, stat: str = 'min'):
        """Normalize data, default is min which keeps values between 0 and 1"""
        max_min_val = self.cleanData.max - self.cleanData.min
        if max_min_val == 0.0:
            max_min_val = 1.0
        if stat == "mean":
            lst = ((val - self.cleanData.mean) / max_min_val for val in self.cleanData.data)
        elif stat == "min":
            lst = ((val - self.cleanData.min) / max_min_val for val in self.cleanData.data)
        elif stat == "median":
            lst = ((val - self.cleanData.median) / max_min_val for val in self.cleanData.data)
        else:
            raise AttributeError('Stat must be (mean, min, median)')
        self.data = _check_type(d=lst, dtype=self.cleanData.dtype)
        self.len = self.data.__len__()
        self.name = self.cleanData.name + "_normalize_" + stat
        return self

    def add_standardize(self, stat: str = "mean"):
        """Standardize data, with a mean of 0 and std of 1"""
        if stat == "mean":
            lst = ((item - self.cleanData.mean) / self.cleanData.std for item in self.cleanData.data)
        elif stat == "median":
            temp_std = (self.cleanData.higher - self.cleanData.median)
            lst = ((item - self.cleanData.median) / temp_std for item in self.cleanData.data)
        self.data = _check_type(d=lst, dtype=self.cleanData.dtype)
        self.len = self.data.__len__()
        self.name = self.cleanData.name + "_standardize_" + stat
        return self

    def add_running(self,  window: int, stat: str = "mean", q: float = 0.50):
        """Calc running statistics"""
        calc = {"min": _min, "max": _max, "mean": _mean, "var": _var, "std": _std, "sum": _sum, "median": _med,
                "mode": _mod, "skew": _skew, "kurt": _kurt, "percentile": _perc}[stat]
        ran = range(window, self.cleanData.len)
        if stat != "percentile":
            self.name = self.cleanData.name + "_running_" + stat
            pre = [calc(data=self.cleanData.data[:window])] * window
            post = [calc(data=self.cleanData.data[i - window:i]) for i in ran]
        else:
            self.name = self.cleanData.name + "_running_" + stat + "_" + str(q)
            pre = [_perc(d=self.cleanData.data[:window], q=q)] * window
            post = [_perc(d=self.cleanData.data[i - window:i], q=q) for i in ran]
        self.data = _check_type(d=pre + post, dtype=self.cleanData.dtype)
        self.len = self.data.__len__()
        return self

    def add_cumulative(self, stat: str = "mean", q: float = 0.75):
        """Calc cumulative statistics"""
        calc = {"min": _min, "max": _max, "mean": _mean, "var": _var, "std": _std, "sum": _sum, "median": _med,
                "mode": _mod, "skew": _skew, "kurt": _kurt, "percentile": _perc}[stat]
        ran = range(1, self.cleanData.len)
        if stat != "percentile":
            self.name = self.cleanData.name + "_running_" + stat + "_" + str(q)
            lst = [0.0] + [calc(data=self.cleanData.data[:i]) for i in ran]
        else:
            self.name = self.cleanData.name + "_running_" + stat
            lst = [0.0] + [calc(data=self.cleanData.data[:i], q=q) for i in ran]
        self.data = _check_type(d=lst, dtype=self.cleanData.dtype)
        self.len = self.data.__len__()
        return self

    def add_log(self, constant: float = .01):
        lst = (math.log(i + constant) for i in self.cleanData.data)
        self.data = _check_type(d=lst, dtype=self.cleanData.dtype)
        self.name = self.cleanData.name + "_log"
        self.len = self.data.__len__()
        return self

    def add_box_cox(self, lam: float = 0.1):
        """Only postive values"""
        """lambda = -1. is a reciprocal transform.
           lambda = -0.5 is a reciprocal square root transform.
           lambda = 0.0 is a log transform.
           lambda = 0.5 is a square root transform.
           lambda = 1.0 is no transform."""
        if self.cleanData is None:
            return self
        lst = ((i ** lam - 1) / lam for i in self.cleanData.data)
        self.data = _check_type(d=lst, dtype=self.cleanData.dtype)
        self.name = self.cleanData.name + "_box_cox_" + str(lam)
        self.len = self.data.__len__()
        return self

    def add_sklearn_box_cox(self, standard: bool = True):
        """Only postive values"""
        arr = power_transform(X=self.cleanData.array(axis=1), method='box-cox', standardize=standard)
        self.data = _check_type(d=(i[0] for i in _mtype(d=arr)), dtype=self.cleanData.dtype)
        self.name = self.cleanData.name + "_sklearn_box_cox"
        self.len = self.data.__len__()
        return self

    def add_sklearn_yeo_johnson(self, standard: bool = True):
        """Postive values and negative values"""
        arr = power_transform(X=self.cleanData.array(axis=1), method='yeo-johnson', standardize=standard)
        self.data = _check_type(d=(i[0] for i in _mtype(d=arr)), dtype=self.cleanData.dtype)
        self.name = self.cleanData.name + "_sklearn_yeo_johnson"
        self.len = self.data.__len__()
        return self

    def add_sklearn_quantile(self, n_quantiles: int = 25, output_distribution: str = 'uniform'):
        """Recommended to not do before splitting"""
        """Also accepts 'normal' """
        arr = quantile_transform(X=self.cleanData.array(axis=1), n_quantiles=n_quantiles,
                                 output_distribution=output_distribution)
        self.data = _check_type(d=(i[0] for i in _mtype(d=arr)), dtype=self.cleanData.dtype)
        self.name = self.cleanData.name + "_sklearn_quantile_" + str(n_quantiles) + "_" + output_distribution
        self.len = self.data.__len__()
        return self

    def add_sklearn_robust_scaling(self, with_centering: bool = True, with_scaling: bool = True,
                                   quantile_range: tuple = (25.0, 75.0)):
        """Recommended to not do before splitting"""
        arr = robust_scale(X=self.cleanData.array(axis=1), with_centering=with_centering, with_scaling=with_scaling,
                           quantile_range=quantile_range)
        self.data = _check_type(d=(i[0] for i in _mtype(d=arr)), dtype=self.cleanData.dtype)
        self.name = self.cleanData.name + "_sklearn_robust"
        self.len = self.data.__len__()
        return self

    def add_constant(self, columns: int = 2, other: Data = None):
        """Adds a column of 1's to the data"""
        if other:
            if other.len != self.cleanData.len:
                if other.name is None:
                    name1 = "First Array"
                else:
                    name1 = other.name
                if self.cleanData.name is None:
                    name2 = "Second Array"
                else:
                    name2 = self.cleanData.name
                raise AttributeError("Lengths of the two Data's are different: " + name1 + " {}, " + name2 + " {})".format((other.len, self.cleanData.len)))
            else:
                arr = np.ones((self.cleanData.len, 2))
                arr[:, 0] = self.cleanData.array(axis=0)
                arr[:, 1] = other.array(axis=0)
        else:
            arr = np.ones((self.cleanData.len, columns))
            arr[:, 1] = self.data
            name1 = 'constant'
            name2 = self.name
        self.data = arr
        self.name = (name1, name2)
        return self

    # Type Change Methods
    def list(self) -> list:
        """Returns a list"""
        if isinstance(self.data, (list, tuple)):
            return list(self.data)
        else:
            raise AttributeError("PreProcess data must be a list or tuple prior to this method call.")

    def tuple(self) -> tuple:
        """Returns a tuple"""
        if isinstance(self.data, (list, tuple)):
            return tuple(self.data)
        else:
            raise AttributeError("PreProcess data must be a list or tuple prior to this method call.")

    def array(self, axis: int = 0) -> np.ndarray:
        """Returns an np.ndarray"""
        if axis == 0:
            return np.array(self.data)
        else:
            return np.array(self.data).reshape(self.len, 1)

    def dataframe(self, index: list = None, name: str = None) -> DataFrame:
        """Returns a pd.DataFrame"""
        if index is None:
            index = range(self.len)

        if len(index) != self.len:
            raise AttributeError("List of index {} not equal to length of data {}.".format(len(index), self.len))

        if name is None and self.name is None:
            raise AttributeError("Need to pass a name.")
        else:
            if name is None:
                name = self.name
        return DataFrame(self.data, columns=[name], index=index)

    def dict(self) -> dict:
        """Returns a dict"""
        if isinstance(self.data, np.ndarray):
            return {self.name[ind]: self.data[:, ind] for ind, val in enumerate(self.data)}
        else:
            return {self.name: self.data}

    def __repr__(self):
        return 'PreProcessData'



"""
Hyperparameter tuning class.

Usage:
 ./models/tuning.py

Author:
 Peter Rigali - 2022-03-19
"""
from dataclasses import dataclass
import numpy as np
from pyjr.classes.model_data import ModelingData
from sklearn.model_selection import cross_val_score, cross_val_predict, learning_curve, permutation_test_score, validation_curve
import warnings
warnings.filterwarnings("ignore")


@dataclass
class Tune:

    __slots__ = ["data", "pred", "model", "cv", "scores", "jobs", "train_sizes"]

    def __init__(self, model, data: ModelingData, jobs: int = 1, cv: int = 10):
        self.data = data
        self.model = model
        self.cv = cv
        self.scores = {}
        self.jobs = jobs

        self.scores['standard'] = tuple(cross_val_score(estimator=self.model, X=self.data.x_train, y=self.data.y_train, cv=self.cv, n_jobs=self.jobs).tolist())
        self.pred = tuple(cross_val_predict(estimator=self.model, X=self.data.x_train, y=self.data.y_train, cv=self.cv, n_jobs=self.jobs).tolist())
        lr = learning_curve(estimator=self.model, X=self.data.x_train, y=self.data.y_train, cv=self.cv, n_jobs=self.jobs)
        self.train_sizes, self.scores['training'], self.scores['testing'] = [tuple(i) for i in lr]
        self.scores['permutation'] = permutation_test_score(estimator=self.model, X=self.data.x_train, y=self.data.y_train, cv=self.cv, n_jobs=self.jobs)

        for key in ['testing', "training"]:
            temp = []
            for val in self.scores[key]:
                if isinstance(val, np.ndarray):
                    temp.append(tuple(val.tolist()))
                else:
                    temp.append(val)
            self.scores[key] = tuple(temp)

    def __repr__(self):
        return "ModelTuning"

    def add_score(self):
        self.scores['standard'] = tuple(cross_val_score(estimator=self.model, X=self.data.x_train, y=self.data.y_train, cv=self.cv, n_jobs=self.jobs).tolist())
        return self

    def add_pred(self):
        self.pred = tuple(cross_val_predict(estimator=self.model, X=self.data.x_train, y=self.data.y_train, cv=self.cv, n_jobs=self.jobs).tolist())
        return self

    def add_learning_curve(self):
        lr = learning_curve(estimator=self.model, X=self.data.x_train, y=self.data.y_train, cv=self.cv, n_jobs=self.jobs)
        self.train_sizes, self.scores['traing'], self.scores['testing'] = [tuple(i) for i in lr]
        return self

    def add_permutation_score(self):
        self.scores['permutation'] = permutation_test_score(estimator=self.model, X=self.data.x_train, y=self.data.y_train, cv=self.cv, n_jobs=self.jobs)
        return self



"""
Dict functions.

Usage:
 ./utils/tools/dic.py

Author:
 Peter Rigali - 2022-03-30
"""
from typing import Union


def _search_dic_values(dic: dict, item: Union[str, int, float]) -> Union[str, float, int]:
    """

    Searches a dict using the values.

    :param dic: Input data.
    :type dic: dict
    :param item: Search item.
    :type item: str, float or int
    :return: Key value connected to the value.
    :rtype: str, float or int
    :example: *None*
    :note: *None*

    """
    return list(dic.keys())[list(dic.values()).index(item)]


"""
General functions.

Usage:
 ./utils/tools/general.py

Author:
 Peter Rigali - 2022-03-30
"""
from typing import Union, List
import numpy as np


def _unique_values(data: Union[list, tuple], count: False):
    """
    Finds unique values from a list.

    :param data: Input data.
    :type data: list or tuple.
    :return: Returns either a list or dict.
    :note: *None*
    """
    if count:
        unique = set(data)
        return {i: data.count(i) for i in unique}
    else:
        return tuple(set(data))


def _add_constant(data: Union[list, tuple, np.ndarray]) -> np.ndarray:
    """Add a column of ones to a list, tuple or np.ndarray"""
    if isinstance(data, (tuple, list)):
        arr = np.ones((data.__len__(), 2))
    elif isinstance(data, np.ndarray):
        arr = np.ones((data.shape[0], 2))
    arr[:, 1] = data
    return arr


def _cent(x_lst: List[float], y_lst: List[float]) -> List[float]:
    """

    Calculate Centroid from x and y value(s).

    :param x_lst: A list of values.
    :type x_lst: List[float]
    :param y_lst: A list of values.
    :type y_lst: List[float]
    :returns: A list of x and y values representing the centriod of two lists.
    :rtype: List[float]
    :example: *None*
    :note: *None*

    """
    return [np.sum(x_lst) / len(x_lst), np.sum(y_lst) / len(y_lst)]


def _dis(c1: List[float], c2: List[float]) -> float:
    """

    Calculate Distance between two centroids.

    :param c1: An x, y coordinate representing a centroid.
    :type c1: List[float]
    :param c2: An x, y coordinate representing a centroid.
    :type c2: List[float]
    :returns: A distance measurement.
    :rtype: float
    :example: *None*
    :note: *None*

    """
    return round(np.sqrt((c1[0] - c2[0]) ** 2 + (c1[1] - c2[1]) ** 2), 4)



"""
ModelData class.

Usage:
 ./utils/model_data.py

Author:
 Peter Rigali - 2022-03-19
"""
from dataclasses import dataclass
from typing import Union, List, Tuple
import numpy as np
import random
from pyjr.classes.data import Data
from pyjr.classes.preprocess_data import PreProcess
from pyjr.utils.tools.math import _sum
from pyjr.utils.tools.clean import _round, _mtype
from pyjr.utils.tools.general import _unique_values
from pyjr.utils.tools.array import _add_column
from pyjr.utils._class_functions import _len, _names
from sklearn.decomposition import PCA, TruncatedSVD


@dataclass
class ModelingData:

    __slots__ = ("x_data", "x_data_names", "len", "y_data", "y_data_name", "x_train", "x_test", "x_valid", "y_train",
                 "y_test", "y_valid", "train_ind", "test_ind", "valid_ind")

    def __init__(self):
        self.x_data = None
        self.x_data_names = []
        self.len = None
        self.y_data = None
        self.y_data_name = None
        self.x_train = None
        self.x_test = None
        self.x_valid = None
        self.y_train = None
        self.y_test = None
        self.y_valid = None
        self.train_ind = None
        self.test_ind = None
        self.valid_ind = None

    def __repr__(self):
        return "ModelingData"

    def add_xdata(self, data: Union[Data, PreProcess]):
        if self.x_data is None:
            if isinstance(data, Data):
                self.x_data = data.array(axis=1)
            else:
                self.x_data = np.array(data.data).reshape(data.len, 1)
            self.len = data.len
            self.x_data_names.append(data.name)
        else:
            if _len(l1=self.len, l2=data.len) is True and _names(n=data.name, n_lst=self.x_data_names) is True:
                self.x_data_names.append(data.name)
                if isinstance(data, Data):
                    self.x_data = _add_column(arr1=self.x_data, arr2=data.array(axis=1))
                else:
                    self.x_data = _add_column(arr1=self.x_data, arr2=np.array(data.data).reshape(data.len, 1))
        return self

    def add_ydata(self, data: Union[Data, PreProcess]):
        if _len(l1=self.len, l2=data.len):
            if isinstance(data, Data):
                self.y_data = data.array(axis=1)
            else:
                self.y_data = np.array(data.data).reshape(data.len, 1)
            self.y_data_name = data.name
            return self

    def add_train_test_validate(self, train: Union[int, float] = 0.70, test: Union[int, float] = 0.20,
                                valid: Union[int, float] = 0.10):

        if isinstance(train, float):
            train = int(np.floor(train * float(self.len)))
        if isinstance(test, float):
            test = int(np.floor(test * float(self.len)))
        if isinstance(valid, float):
            valid = int(np.floor(valid * float(self.len)))

        if train + test + valid > self.len:
            raise AttributeError("Train, test, valid add up to more than the length of the data.")

        vals = tuple(range(self.len))
        self.train_ind = random.sample(vals, train)
        train_dic = {i: True for i in self.train_ind}
        self.x_train = self.x_data[self.train_ind]
        self.y_train = self.y_data[self.train_ind]

        vals = [val for val in vals if val not in train_dic]
        self.test_ind = random.sample(vals, test)
        test_dic = {i: True for i in self.test_ind}
        self.x_test = self.x_data[self.test_ind]
        self.y_test = self.y_data[self.test_ind]

        if valid != 0:
            self.valid_ind = [val for val in vals if val not in test_dic]
            self.x_valid = self.x_data[self.valid_ind]
            self.y_valid = self.y_data[self.valid_ind]

        return self.x_train, self.y_train, self.x_test, self.y_test, self.x_valid, self.y_valid

    def add_multiple_xdata(self, data: Union[List[Union[Data, PreProcess]], Tuple[Union[Data, PreProcess]]]):
        """Calls the add_xdata method to add multiple Data's"""
        for i in data:
            self.add_xdata(data=i)
        return self

    def get_balance(self):
        """Returns the data balance for Y_data between train, test, and valid"""
        train_lst = _mtype(d=self.y_train.reshape(1, self.y_train.shape[0])[0])
        test_lst = _mtype(d=self.y_test.reshape(1, self.y_test.shape[0])[0])
        dic = {"train": _unique_values(data=train_lst, count=True),
               "test": _unique_values(data=test_lst, count=True)}

        if self.y_valid is not None:
            valid_lst = _mtype(d=self.y_valid.reshape(1, self.y_valid.shape[0])[0])
            dic["valid"] = _unique_values(data=valid_lst, count=True)

        final_dic = {i: {} for i in dic.keys()}
        for key, val in dic.items():
            for key1, val1 in val.items():
                final_dic[key][key1] = _round(d=val[key1] / _sum(d=_mtype(d=val.values())), v=100, r=True)
        return final_dic

    def add_pca(self, n_com: int = 2):
        # 'mle' is default
        pca = PCA(n_components=n_com)
        pca.fit(self.x_data)
        self.x_data = pca.transform(self.x_data)
        self.x_data_names = ["pca_col_" + str(i) for i in range(pca.explained_variance_.size)]
        return pca.explained_variance_ratio_

    def add_truncatedsvd(self, n_com: int = 2):
        svd = TruncatedSVD(n_components=n_com)
        svd.fit_transform(self.x_data)
        self.x_data = svd.transform(self.x_data)
        self.x_data_names = ["truncated_col_" + str(i) for i in range(svd.explained_variance_.size)]
        return svd.explained_variance_ratio_

    def add_ohe_data(self, data: Data):
        unique = _unique_values(data=data.data, count=False)
        for ind in range(len(unique)):
            self.add_xdata(data=Data(data=[1.0 if ind == val else 0.0 for val in data.data],
                                     name=str(unique[ind]) + "_ohe"))
        return self



"""
Array functions.

Usage:
 ./utils/tools/array.py

Author:
 Peter Rigali - 2022-03-30
"""
from typing import Optional
import numpy as np


def _stack(x_arr: np.ndarray, y_arr: np.ndarray, multi: Optional[bool] = False) -> np.ndarray:
    """

    Stacks x_arr and y_arr.

    :param x_arr: An array to stack.
    :type x_arr: np.ndarray
    :param y_arr: An array to stack.
    :type y_arr: np.ndarray
    :param multi: If True, will stack based on multiple x_arr columns, default is False. *Optional*
    :type multi: bool
    :return: Array with a x column and a y column
    :rtype: np.ndarray
    :example: *None*
    :note: *None*

    """
    lst = []
    if multi:
        for i in range((x_arr.shape[1])):
            lst.append(np.vstack([x_arr[:, i].ravel(), y_arr[:, i].ravel()]).T)
        return np.array(lst)
    else:
        lst = np.vstack([x_arr.ravel(), y_arr.ravel()]).T
    return np.where(np.isnan(lst), 0, lst)


def _add_column(arr1: np.ndarray, arr2: np.ndarray) -> np.ndarray:
    """Adds a column-arr to an arr"""
    new_arr = np.ones((arr1.shape[0], arr1.shape[1] + 1))
    for i in range(arr1.shape[1]):
        new_arr[:, i] = arr1[:, i]
    new_arr[:, new_arr.shape[1] - 1] = arr2[:, 0]
    return new_arr



"""
CleanData class.

Usage:
 ./utils/data.py

Author:
 Peter Rigali - 2022-03-10
"""
from dataclasses import dataclass
from pandas import DataFrame
import numpy as np
from scipy.stats import kstest, normaltest, shapiro
from pyjr.utils.tools.math import _min, _max, _mean, _var, _std, _sum, _med, _mod, _skew, _kurt, _perc
from pyjr.utils.tools.math import _percs, _range
from pyjr.utils.tools.clean import _nan, _prep, _mtype
from pyjr.utils.tools.general import _unique_values


@dataclass
class Data:
    """

    Builds CleanData Class. Used for analysis of data.

    :param data: Input data.
    :type data:
    :param name: Input data name.
    :type name: str.
    :param na_handling: Desired Nan value handling method. {zero, mu, std, median}
    :type na_handling: str.
    :param dtype: Desired type to fit data to.
    :type dtype: str.
    :param cap_zero: Whether to cap the value at zero.
    :type cap_zero: bool.
    :param std_value: Desired Standard Deviation to use.
    :type std_value: int.
    :param ddof: Desired Degrees of Freedom.
    :type ddof: int.
    :param q_lst: List of columns to find the Quantile. *Optional*
    :type q_lst: list of floats.
    :example: *None*
    :note: *None*

    """

    __slots__ = ("name", "data", "len", "unique", "dtype", "mean", "median", "mode", "var", "std", "lower", "higher",
                 "min", "max", "sum", "skew", "kurt", "per", "rang", "distribution", "na")

    def __init__(self, data = None, name: str = None, na_handling: str = 'none',
                 dtype: str = 'float', cap_zero: bool = True, std_value: int = 3, median_value: float = 0.023,
                 ddof: int = 1, q_lst: tuple = (0.159, 0.841), stats: bool = True,
                 distribution: bool = False, unique: bool = False):

        self.name = None
        if name:
            self.name = name
        self.data = _prep(d=data, dtype=dtype, na=na_handling)
        self.unique = None
        if unique:
            self.unique = _unique_values(data=self.data, count=False)

        self.len = self.data.__len__()
        self.dtype = dtype
        if stats and dtype in {"float": True, "int": True}:
            self.mean = _mean(d=self.data)
            self.median = _med(d=self.data)
            self.mode = _mod(d=self.data)
            self.var = _var(d=self.data, dof=ddof)
            self.std = _std(d=self.data, dof=ddof)
            self.lower, self.higher = _percs(d=self.data, q_lst=q_lst)
            self.min = _min(d=self.data)
            self.max = _max(d=self.data)
            self.sum = _sum(d=self.data)
            self.skew = _skew(d=self.data)
            self.kurt = _kurt(d=self.data)
            self.per = _perc(d=self.data, q=0.75)
            self.rang = _range(d=self.data)
        else:
            self.mean, self.median, self.mode, self.var, self.std = None, None, None, None, None
            self.lower, self.higher, self.min, self.max, self.sum = None, None, None, None, None
            self.skew, self.kurt, self.per, self.rang = None, None, None, None

        self.na = None
        if self.data.__len__() != data.__len__():
            tup = _mtype(d=data, dtype='tuple')
            _na_ind_lst = [ind for ind, val in enumerate(tup) if _nan(v=val) == True]
            if _na_ind_lst.__len__() > 0:
                _percent_na = _na_ind_lst.__len__() / self.len
                self.na = {"index": _na_ind_lst, "percent": _percent_na}

        self.distribution = None
        if distribution:
            self.distribution, count = {'Kolmogorov-Smirnov': kstest(self.data, 'norm')[1],
                                        'DAgostino': normaltest(self.data)[1],
                                        'Shapiro-Wilk': shapiro(self.data)[1],
                                        'noraml': False}, 0
            for test_name in ('Kolmogorov-Smirnov', "DAgostino", 'Shapiro-Wilk'):
                if self.distribution[test_name] >= .05:
                    count += 1
            if count == 0:
                self.distribution['normal'] = True

    def add_percentile(self, q: float) -> float:
        self.per = _perc(d=self.data, q=q)
        return self.per

    # Type Change Methods
    def astype(self, dtype):
        self.data = tuple([dtype(i) for i in self.data])
        self.dtype = dtype.__str__()
        return self

    def list(self) -> list:
        """Returns a list"""
        return list(self.data)

    def tuple(self) -> tuple:
        """Returns a tuple"""
        return tuple(self.data)

    def array(self, axis: int = 0) -> np.ndarray:
        """Returns an np.ndarray"""
        if axis == 0:
            return np.array(self.data)
        else:
            return np.array(self.data).reshape(self.len, 1)

    def dataframe(self, index: list = None, name: str = None) -> DataFrame:
        """Returns a pd.DataFrame"""
        if index is None:
            index = range(self.len)

        if len(index) != self.len:
            raise AttributeError("List of index {} not equal to length of data {}.".format(len(index), self.len))

        if name is None and self.name is None:
            raise AttributeError("Need to pass a name.")
        else:
            if name is None:
                name = self.name
        return DataFrame(self.data, columns=[name], index=index)

    def dict(self) -> dict:
        """Returns a dict"""
        return {self.name: self.data}

    def __repr__(self):
        return 'CleanData'



"""
Barchart plot class.

Usage:
 ./plot/table.py

Author:
 Peter Rigali - 2022-03-10
"""
from dataclasses import dataclass
from typing import List, Union, Tuple
import matplotlib.pyplot as plt
import pandas as pd
from pyjr.classes.data import Data
from pyjr.classes.preprocess_data import PreProcess
from pyjr.utils.tools.clean import _mtype
from pyjr.utils.tools.math import _mean, _std, _sum, _med


@dataclass
class Bar:
    """

    Builds Bar Chart Class. Used for plotting data.

    :param data: Input data.
    :type data: Either Data(pyjr class), PreProcess(pyjr class), list of the latter or a dict, pd.DataFrame.
    :param value_string: Input data conversion, {sum, mean, median, std}. *Default is sum.*
    :type value_string: str.
    :param label_lst: Columns or names to focus on.
    :type label_lst: list or tuple of str's.
    :param vert_hor: True is vertical, False is horizontal. *Default is True.*
    :type vert_hor: bool.
    :param xlabel: Plot X axis label.
    :type xlabel: str
    :param xlabel_size: Plot X axis font size. *Default is medium*
    :type xlabel_size: str
    :param ylabel: Plot Y axis label.
    :type ylabel: str
    :param ylabel_size: Plot Y axis font size. *Default is medium*
    :type ylabel_size: str
    :param title: Title of plt.
    :type title: str
    :param title_size: Size of title font. *Default is xx-large*
    :type title_size: str
    :param limit: Selection of rows to include.
    :type limit: list or tuple of int's.
    :param include_mu: Include the mean of the values. *Default is True*
    :type include_mu: bool
    :param mu_color: Color of mu data. *Default is red.*
    :type mu_color: str
    :param color_lst: List or tuple of colors to use in plot.
    :type color_lst: List or tuple
    :param grid: Whether to include a grid. *Default is True.*
    :type grid: bool
    :param grid_alpha: Transparency of grid lines. *Default is 0.75'*
    :type grid_alpha: float
    :param grid_lineweight: Width of gridlines. *Default is 0.5.*
    :type grid_lineweight: float
    :param grid_dash_sequence: Tuple of spaces and lines in grid. *Default is (1, 3)*
    :type grid_dash_sequence: tuple
    :param fig_size: Width and height of figure. *Default is (10, 7).*
    :type fig_size: tuple
    :param show: Whether to print the plot. *Default is False.*
    :type show: bool
    :example: *None*
    :note:
        fonts can be: {'xx-small', 'x-small', 'small', 'medium', 'large', 'x-large', 'xx-large'}
        location can be: {'best', 'upper right', 'upper left', 'lower left', 'lower right', 'right', 'center left',
                          'center right', 'lower center', 'upper center', 'center'}

    """
    __slots__ = "ax"

    def __init__(self,
                 data: Union[pd.DataFrame, Data, PreProcess, List[Union[Data, PreProcess]], dict],
                 value_string: str = 'sum',
                 label_lst: Union[List[str], Tuple[str]] = None,
                 vert_hor: bool = True,
                 xlabel: str = 'Names',
                 xlabel_size: str = 'medium',
                 ylabel: str = 'Values',
                 ylabel_size: str = 'medium',
                 title: str = 'Bar Chart',
                 title_size: str = 'xx-large',
                 limit: Union[List[int], Tuple[int]] = None,
                 include_mu: bool = False,
                 mu_color: str = 'r',
                 color_lst: Union[list, tuple] = None,
                 grid: bool = True,
                 grid_alpha: float = 0.75,
                 grid_lineweight: float = 0.5,
                 grid_dash_sequence: tuple = (1, 3),
                 fig_size: tuple = (10, 7),
                 show: bool = False,
                 ):
        # Parse input data
        dic = False
        if isinstance(data, (Data, PreProcess)):
            if label_lst is None:
                label_lst = _mtype(d=data.name, dtype='list')
            data = data.dataframe()
        elif isinstance(data, pd.DataFrame):
            if label_lst is None:
                label_lst = _mtype(d=data.columns, dtype='list')
        elif isinstance(data, list):
            _dic = {}
            for d in data:
                if isinstance(d.name, (list, tuple)):
                    for ind, val in enumerate(d.name):
                        _dic[val.name] = val.data[:, ind]
                else:
                    _dic[d.name] = d.data
            data = pd.DataFrame.from_dict(_dic)
            label_lst = _mtype(d=data.columns, dtype='list')
        elif isinstance(data, dict):
            value_lst = _mtype(d=data.values(), dtype='list')
            label_lst = _mtype(d=data.keys(), dtype='list')
            dic = True

        if dic is False:
            if limit:
                data = data[limit[0]:limit[1]]
            # Get values
            value_lst = []
            for key in label_lst:
                val = _mtype(d=data[key], dtype='list')
                if value_string == 'sum':
                    val = _sum(d=val)
                elif value_string == 'mean':
                    val = _mean(d=val)
                elif value_string == 'median':
                    val = _med(d=val)
                elif value_string == 'std':
                    val = _std(d=val)
                value_lst.append(val)

        # Get colors
        if color_lst is None:
            color_lst = ['tab:orange' for i in range(label_lst.__len__())]
        elif color_lst == 'gradient':
            color_lst = [plt.get_cmap('viridis')(1. * i / label_lst.__len__()) for i in range(label_lst.__len__())]
        elif isinstance(color_lst, str) and color_lst != 'gradient':
            color_lst = [color_lst for i in range(label_lst.__len__())]

        if include_mu:
            value_lst.append(_mean(value_lst))
            label_lst.append('mu')
            color_lst.append(mu_color)

        # Start plot
        fig, ax = plt.subplots(figsize=fig_size)

        # Plot values
        if vert_hor:
            ax.bar(label_lst, value_lst, color=color_lst)
        else:
            ax.barh(label_lst, value_lst, color=color_lst)
            if ylabel == 'Values':
                xlabel = 'Values'
                ylabel = 'Names'

        plt.ylabel(ylabel, fontsize=ylabel_size)
        plt.xlabel(xlabel, fontsize=xlabel_size)
        plt.title(title, fontsize=title_size)

        # Add grid
        if grid:
            ax.grid(alpha=grid_alpha, linestyle=(0, grid_dash_sequence), linewidth=grid_lineweight)

        self.ax = ax

        if show:
            plt.show()

    def __repr__(self):
        return 'Bar Chart Plot'
